"""
Module for importing remote and local files into CARTO

.. module:: carto.field_import
   :platform: Unix, Windows
   :synopsis: Module for importing remote and local files into CARTO

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from pyrestcli.fields import IntegerField, CharField, BooleanField

from .exceptions import CartoException
from .resources import AsyncResource, Manager
from .paginators import CartoPaginator


API_VERSION = "v1"
API_ENDPOINT = 'api/{api_version}/imports/'


class FileImportJob(AsyncResource):
    """
    This class provides support for one-time uploading and importing of
    remote and local files into CARTO
    """
    item_queue_id = CharField()
    id = CharField()
    user_id = CharField()
    table_id = CharField()
    data_type = CharField()
    table_name = CharField()
    state = CharField()
    error_code = IntegerField()
    queue_id = CharField()
    tables_created_count = IntegerField()
    synchronization_id = CharField()
    type_guessing = BooleanField()
    quoted_fields_guessing = BooleanField()
    content_guessing = BooleanField()
    create_visualization = BooleanField()
    visualization_id = CharField()
    user_defined_limits = CharField()
    get_error_text = None
    display_name = CharField()
    success = BooleanField()
    warnings = None
    is_raster = BooleanField()

    class Meta:
        collection_endpoint = API_ENDPOINT.format(api_version=API_VERSION)
        id_field = "item_queue_id"

    def __init__(self, archive, auth_client):
        """

        :param auth_client: Client to make authorized requests
                            (currently only APIKeyAuthClient is supported)
        :param archive: archive can be a pointer to a remote location, a path
                     to a local file or a StringIO object
        :type auth_client: :class:`carto.auth.APIKeyAuthClient`
        :type archive: str

        :return:
        """
        self.file = None
        self.files = None
        if archive is not None:
            if hasattr(archive, "startswith") and archive.startswith("http"):
                self.file = archive
            else:
                self.files = {'file': self.__open(archive, 'rb')}

        super(FileImportJob, self).__init__(auth_client)

    def run(self, **import_params):
        """
        Actually creates the import job on the CARTO server

        :param import_params: To be send to the Import API, see CARTO's docs
                                on Import API for an updated list of accepted
                                params
        :type import_params: kwargs

        :return:

        .. note:: The import job is asynchronous, so you should take care of the progression, by calling the :func:`carto.resources.AsyncResource.refresh` method and check the import job :py:attr:`~state` attribute. See :func:`carto.datasets.DatasetManager.create` for a unified method to import files into CARTO
        """
        if self.file:
            import_params["url"] = self.file

        self.id_field = "id"

        if "connection" in import_params:
            self.fields.append("connector")
            self.update_from_dict(import_params["connection"])
            self.save(force_create=True)
        else:
            super(FileImportJob, self).run(params=import_params,
                                           files=self.files)

    def __open(self, name, mode):
        if hasattr(name, "read"):
            return name
        else:
            return open(name, mode)


class FileImportJobManager(Manager):
    resource_class = FileImportJob
    json_collection_attribute = "imports"
    paginator_class = CartoPaginator

    def filter(self):
        """
        Get a filtered list of file imports

        :return: A list of file imports, with only the id set (you need to
                refresh them if you want all the attributes to be filled in)
        :rtype: list of :class:`carto.file_import.FileImportJob`

        :raise: CartoException
        """
        try:
            response = self.send(self.get_collection_endpoint(), "get")
            if self.json_collection_attribute is not None:
                resource_ids = self.client.get_response_data(
                    response,
                    self.Meta.parse_json)[self.json_collection_attribute]
            else:
                resource_ids = self.client.get_response_data(
                    response, self.Meta.parse_json)
        except Exception as e:
            raise CartoException(e)

        resources = []

        for resource_id in resource_ids:
            try:
                resource = self.resource_class(self.client)
            except (ValueError, TypeError):
                continue
            else:
                setattr(resource, resource.Meta.id_field, resource_id)
                resources.append(resource)

        return resources

    def create(self, archive, **kwargs):
        """
        Creates a file import on the server

        :param archive: archive can be a pointer to a remote location, a path
         to a local file or a StringIO object
        :param kwargs: Attributes (field names and values) of the new resource

        :type archive: str
        :type kwargs: kwargs

        :return: The :class:`carto.file_import.FileImportJob`
        """
        resource = self.resource_class(archive, self.client)
        resource.update_from_dict(kwargs)
        resource.save(force_create=True)

        return resource



"""
Module for working with users

.. module:: carto.users
   :platform: Unix, Windows
   :synopsis: Module for working with users

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from gettext import gettext as _
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from pyrestcli.fields import IntegerField, CharField

from .exceptions import CartoException
from .paginators import CartoPaginator
from .resources import Manager, WarnResource


API_VERSION = "v1"
API_ENDPOINT = "api/{api_version}/organization/{organization}/users/"
API_NONORG_ENDPOINT = "api/{api_version}/users/"


class User(WarnResource):
    """
    Represents an enterprise CARTO user, i.e. a user that belongs to an
    organization

    Currently, CARTO's user API only supports enterprise users.

    .. warning:: Non-public API. It may change with no previous notice
    """
    username = CharField()
    email = CharField()
    avatar_url = CharField()
    base_url = CharField()
    quota_in_bytes = IntegerField()
    db_size_in_bytes = IntegerField()
    table_count = IntegerField()
    public_visualization_count = IntegerField()
    all_visualization_count = IntegerField()
    password = CharField()
    soft_geocoding_limit = IntegerField()

    class Meta:
        id_field = "username"
        name_field = "username"

    def __init__(self, auth_client):
        """
        Sets the collection endpoint dynamically, based on provided
        organization (see api/auth.py)
        :param auth_client: Auth client
        """
        if auth_client.organization is None:
            self._api_endpoint = API_NONORG_ENDPOINT
        else:
            self._api_endpoint = API_ENDPOINT

        super(User, self).__init__(auth_client)

    def get_collection_endpoint(self):
        """
        """
        return self._api_endpoint.format(api_version=API_VERSION,
                                   organization=self.client.organization)

    def get_resource_endpoint(self):
        """
        """
        resource_id = getattr(self, self.Meta.id_field, None)
        if resource_id is None:
            return None
        return urljoin(self.get_collection_endpoint(), str(resource_id))


class UserManager(Manager):
    """
    Manager for the User class.

    .. warning:: Non-public API. It may change with no previous notice
    """
    resource_class = User
    paginator_class = CartoPaginator

    def __init__(self, auth_client):
        """
        :param auth_client: Client to make (non)authorized requests
        :return:
        """
        super(UserManager, self).__init__(auth_client)
        if auth_client.organization is None:
            self._api_endpoint = API_NONORG_ENDPOINT
        else:
            self._api_endpoint = API_ENDPOINT

    def filter(self, **search_args):
        """
        Should get all the current users from CARTO, but this is currently not
        supported by the API
        """
        raise CartoException(_("Retrieving user list is not currently \
                               supported by the API"))

    def get_collection_endpoint(self):
        """
        """
        return self._api_endpoint.format(api_version=API_VERSION,
                                   organization=self.client.organization)

    def get_resource_endpoint(self, resource_id):
        """
        """
        if resource_id is None:
            return None
        return urljoin(self.get_collection_endpoint(), str(resource_id))



"""
Module for the IMPORT API with sync tables

.. module:: carto.sync_tables
   :platform: Unix, Windows
   :synopsis: Module for the IMPORT API with sync tables

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from pyrestcli.fields import IntegerField, CharField, BooleanField, \
    DateTimeField

from .exceptions import CartoException
from .resources import AsyncResource, Manager
from .paginators import CartoPaginator


API_VERSION = "v1"
API_ENDPOINT = 'api/{api_version}/synchronizations/'
API_FORCE_SYNC_SUFFIX = 'sync_now'


class SyncTableJob(AsyncResource):
    """
    This class provides support for creating Sync Tables into CARTO
    """
    id = CharField()
    name = CharField()
    interval = IntegerField()
    url = CharField()
    state = CharField()
    created_at = DateTimeField()
    updated_at = DateTimeField()
    run_at = DateTimeField()
    retried_times = IntegerField()
    log_id = CharField()
    error_code = IntegerField()
    error_message = CharField()
    ran_at = DateTimeField()
    modified_at = DateTimeField()
    etag = CharField()
    checksum = CharField()
    user_id = CharField()
    service_name = CharField()
    service_item_id = CharField()
    success = BooleanField()
    type_guessing = BooleanField()
    quoted_fields_guessing = BooleanField()
    content_guessing = BooleanField()
    visualization_id = CharField()
    from_external_source = BooleanField()
    synchronization_id = CharField()
    enqueued = BooleanField()

    class Meta:
        collection_endpoint = API_ENDPOINT.format(api_version=API_VERSION)

    def __init__(self, url, interval, auth_client):
        """
        :param url: Remote URL for the file
        :param interval: Number of seconds between update intervals (>=900)
        :param client: Client to make authorized requests (currently only
                        APIKeyAuthClient is supported)
        :type url: str
        :type interval: int
        :type client: :class:`carto.auth.APIKeyAuthClient`

        :return:
        """
        self.url = url
        self.interval = interval

        super(SyncTableJob, self).__init__(auth_client)

    def run(self, **import_params):
        """
        Actually creates the job import on the CARTO server

        :param import_params: To be send to the Import API, see CARTO's docs
                                on Import API for an updated list of accepted
                                params
        :type import_params: kwargs

        :return:

        .. note:: The sync table job is asynchronous, so you should take care of the progression, by calling the :func:`carto.resources.AsyncResource.refresh` method and check the import job :py:attr:`~state` attribute. See :func:`carto.datasets.DatasetManager.create` for a unified method to import files into CARTO
        """
        import_params["url"] = self.url
        import_params["interval"] = self.interval

        if "connection" in import_params:
            self.fields.append("connector")
            import_params["connection"]["interval"] = self.interval
            self.update_from_dict(import_params["connection"])
            self.save(force_create=True)
        else:
            return super(SyncTableJob, self).run(params=import_params)

    def get_force_sync_endpoint(self):
        """
        Get the relative path to the specific API resource

        :return: Relative path to the resource

        :raise: CartoException
        """
        return urljoin(self.get_resource_endpoint(), API_FORCE_SYNC_SUFFIX)

    def force_sync(self):
        """
        Forces to sync the SyncTableJob

        :return:

        :raise: CartoException
        """
        try:
            self.send(self.get_resource_endpoint(), "put")
        except Exception as e:
            raise CartoException(e)


class SyncTableJobManager(Manager):
    """
    Manager for the SyncTableJob class
    """
    resource_class = SyncTableJob
    json_collection_attribute = "synchronizations"
    paginator_class = CartoPaginator

    def create(self, url, interval, **kwargs):
        """
        Create a sync table on the server

        :param url: URL can be a pointer to a remote location or a path to a
                    local file
        :param interval: Sync interval in seconds
        :param kwargs: Attributes (field names and values) of the new resource
        :type url: str
        :type interval: int
        :type kwargs: kwargs

        :return: SyncTableJob
        """
        resource = self.resource_class(url, interval, self.client)
        resource.update_from_dict(kwargs)
        resource.save(force_create=True)

        return resource



"""
Module for the SQL API

.. module:: carto.sql
   :platform: Unix, Windows
   :synopsis: Module for the SQL API

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

import zlib
import struct

from .exceptions import CartoException
from requests import HTTPError

SQL_API_URL = 'api/{api_version}/sql'
SQL_BATCH_API_URL = 'api/{api_version}/sql/job/'

MAX_GET_QUERY_LEN = 1024

# The chunk size should be a multiple of the filesystem/buffers block
# size. Big values can cause resource starvation and OTOH small values
# incur in some protocol overhead. Typical linux block size is 4 KB.
DEFAULT_CHUNK_SIZE = 8 * 1024  # 8 KB provides good results in practice

# The compression level of gzip/zlib ranges from 1 (fastest, least
# compression) to 9 (slowest, most compression).  In our performance
# tests, we determined that the most efficient way to transmit data
# end-to-end is to use compression levels 1 or 2 (compressing in the
# client, transmiting through network, decompressing and loading in
# the DB). Those levels are also gentle with platform CPU usage.
DEFAULT_COMPRESSION_LEVEL = 1


class SQLClient(object):
    """
    Allows you to send requests to CARTO's SQL API
    """
    def __init__(self, auth_client, api_version='v2'):
        """
        :param auth_client: Auth client to make authorized requests, such as
                            APIKeyAuthClient
        :param api_version: Current version is 'v2'. 'v1' can be used to avoid
                            caching, but it's not guaranteed to work
        :type auth_client: :class:`carto.auth.APIKeyAuthClient`
        :type api_version: str

        :return:
        """
        self.auth_client = auth_client
        self.api_url = SQL_API_URL.format(api_version=api_version)

        self.api_key = getattr(self.auth_client, 'api_key', None)
        self.username = getattr(self.auth_client, 'username', None)
        self.base_url = self.auth_client.base_url

    def send(self, sql, parse_json=True, do_post=True, format=None, **request_args):
        """
        Executes SQL query in a CARTO server

        :param sql: The SQL
        :param parse_json: Set it to False if you want raw reponse
        :param do_post: Set it to True to force post request
        :param format: Any of the data export formats allowed by CARTO's
                        SQL API
        :param request_args: Additional parameters to send with the request
        :type sql: str
        :type parse_json: boolean
        :type do_post: boolean
        :type format: str
        :type request_args: dictionary

        :return: response data, either as json or as a regular
                    response.content object
        :rtype: object

        :raise: CartoException
        """
        try:
            params = {'q': sql}
            if format:
                params['format'] = format
                if format not in ['json', 'geojson']:
                    parse_json = False

            if request_args is not None:
                for attr in request_args:
                    params[attr] = request_args[attr]

            if len(sql) < MAX_GET_QUERY_LEN and do_post is False:
                resp = self.auth_client.send(self.api_url,
                                             'GET',
                                             params=params)
            else:
                resp = self.auth_client.send(self.api_url, 'POST', data=params)

            return self.auth_client.get_response_data(resp, parse_json)
        except Exception as e:
            raise CartoException(e)


class BatchSQLClient(object):
    """
    Allows you to send requests to CARTO's Batch SQL API
    """
    def __init__(self, client, api_version='v2'):
        """
        :param client: Auth client to make authorized requests, such as
                        APIKeyAuthClient
        :param api_version: Current version is 'v2'. 'v1' can be used to avoid
                            caching, but it's not guaranteed to work
        :type auth_client: :class:`carto.auth.APIKeyAuthClient`
        :type api_version: str

        :return:
        """
        self.client = client
        self.api_url = SQL_BATCH_API_URL.format(api_version=api_version)
        self.api_key = self.client.api_key \
            if hasattr(self.client, "api_key") else None

    def update_from_dict(self, data_dict):
        """
        :param data_dict: Dictionary to be mapped into object attributes
        :type data_dict: dict

        :return:
        """
        for k, v in data_dict.items():
            setattr(self, k, v)
        if "item_queue_id" in data_dict:
            self.id = data_dict["item_queue_id"]

    def send(self, url, http_method, json_body=None, http_header=None):
        """
        Executes Batch SQL query in a CARTO server

        :param url: Endpoint url
        :param http_method: The method used to make the request to the API
        :param json_body: The information that needs to be sent, by default
                            is set to None
        :param http_header: The header used to make write requests to the API,
                            by default is none
        :type url: str
        :type http_method: str
        :type json_body: dict
        :type http_header: str

        :return: Response data, either as json or as a regular response.content
                object
        :rtype: object

        :raise: CartoException
        """
        try:
            data = self.client.send(url,
                                    http_method=http_method,
                                    headers=http_header,
                                    json=json_body)
            data_json = self.client.get_response_data(data)
        except Exception as e:
            raise CartoException(e)
        return data_json

    def create(self, sql_query):
        """
        Creates a new batch SQL query.

        Batch SQL jobs are asynchronous, once created you should call
        :func:`carto.sql.BatchSQLClient.read` method given the `job_id`
        to retrieve the state of the batch query

        :param sql_query: The SQL query to be used
        :type sql_query: str or list of str

        :return: Response data, either as json or as a regular response.content
                    object
        :rtype: object

        :raise: CartoException
        """
        header = {'content-type': 'application/json'}
        data = self.send(self.api_url,
                         http_method="POST",
                         json_body={"query": sql_query},
                         http_header=header)
        return data

    def read(self, job_id):
        """
        Reads the information for a specific Batch API request

        :param job_id: The id of the job to be read from
        :type job_id: str

        :return: Response data, either as json or as a regular response.content
                    object
        :rtype: object

        :raise: CartoException
        """
        data = self.send(self.api_url + job_id, http_method="GET")
        return data

    def update(self, job_id, sql_query):
        """
        Updates the sql query of a specific job

        :param job_id: The id of the job to be updated
        :param sql_query: The new SQL query for the job
        :type job_id: str
        :type sql_query: str

        :return: Response data, either as json or as a regular response.content
                    object
        :rtype: object

        :raise: CartoException
        """
        header = {'content-type': 'application/json'}
        data = self.send(self.api_url + job_id,
                         http_method="PUT",
                         json_body={"query": sql_query},
                         http_header=header)
        return data

    def cancel(self, job_id):
        """
        Cancels a job

        :param job_id: The id of the job to be cancelled
        :type job_id: str

        :return: A status code depending on whether the cancel request was
                    successful
        :rtype: str

        :raise CartoException:
        """
        try:
            confirmation = self.send(self.api_url + job_id, http_method="DELETE")
        except CartoException as e:
            if 'Cannot set status from done to cancelled' in e.args[0].args[0]:
                return 'done'
            else:
                raise e
        return confirmation['status']


class CopySQLClient(object):
    """
    Allows to use the PostgreSQL COPY command for efficient streaming
    of data to and from CARTO.
    """

    def __init__(self, client, api_version='v2'):
        """
        :param client: Auth client to make authorized requests, such as
                        APIKeyAuthClient
        :param api_version: Current version is 'v2'. 'v1' can be used to avoid
                            caching, but it's not guaranteed to work
        :type auth_client: :class:`carto.auth.APIKeyAuthClient`
        :type api_version: str

        :return:
        """
        self.client = client
        self.api_url = SQL_API_URL.format(api_version=api_version)
        self.api_key = self.client.api_key \
            if hasattr(self.client, "api_key") else None

    def _read_in_chunks(self, file_object, chunk_size=DEFAULT_CHUNK_SIZE):
        while True:
            data = file_object.read(chunk_size)
            if not data:
                break
            yield data

    def _compress_chunks(self, chunk_generator, compression_level):
        zlib_mode = 16 + zlib.MAX_WBITS
        compressor = zlib.compressobj(compression_level,
                                      zlib.DEFLATED,
                                      zlib_mode)
        for chunk in chunk_generator:
            compressed_chunk = compressor.compress(chunk)
            if len(compressed_chunk) > 0:
                yield compressed_chunk
        yield compressor.flush()

    def copyfrom(self, query, iterable_data, compress=True,
                 compression_level=DEFAULT_COMPRESSION_LEVEL):
        """
        Gets data from an iterable object into a table

        :param query: The "COPY table_name [(column_name[, ...])]
                           FROM STDIN [WITH(option[,...])]" query to execute
        :type query: str

        :param iterable_data: An object that can be iterated
                              to retrieve the data
        :type iterable_data: object

        :return: Response data as json
        :rtype: str

        :raise CartoException:
        """
        url = self.api_url + '/copyfrom'
        headers = {
            'Content-Type': 'application/octet-stream',
            'Transfer-Encoding': 'chunked'
        }
        params = {'api_key': self.api_key, 'q': query}

        if compress:
            headers['Content-Encoding'] = 'gzip'
            _iterable_data = self._compress_chunks(iterable_data,
                                                   compression_level)
        else:
            _iterable_data = iterable_data

        try:
            response = self.client.send(url,
                                        http_method='POST',
                                        params=params,
                                        data=_iterable_data,
                                        headers=headers,
                                        stream=True)
            response_json = self.client.get_response_data(response)
        except Exception as e:
            raise CartoException(e)

        return response_json

    def copyfrom_file_object(self, query, file_object, compress=True,
                             compression_level=DEFAULT_COMPRESSION_LEVEL):
        """
        Gets data from a readable file object into a table

        :param query: The "COPY table_name [(column_name[, ...])]
                           FROM STDIN [WITH(option[,...])]" query to execute
        :type query: str

        :param file_object: A file-like object.
                            Normally the return value of open('file.ext', 'rb')
        :type file_object: file

        :return: Response data as json
        :rtype: str

        :raise CartoException:
        """
        chunk_generator = self._read_in_chunks(file_object)
        return self.copyfrom(query, chunk_generator, compress,
                             compression_level)

    def copyfrom_file_path(self, query, path, compress=True,
                           compression_level=DEFAULT_COMPRESSION_LEVEL):
        """
        Gets data from a readable file into a table

        :param query: The "COPY table_name [(column_name[, ...])]
                           FROM STDIN [WITH(option[,...])]" query to execute
        :type query: str

        :param path: A path to a file
        :type path: str

        :return: Response data as json
        :rtype: str

        :raise CartoException:
        """
        with open(path, 'rb') as f:
            result = self.copyfrom_file_object(query, f, compress,
                                               compression_level)
        return result

    def copyto(self, query):
        """
        Gets data from a table into a Response object that can be iterated

        :param query: The "COPY { table_name [(column_name[, ...])] | (query) }
                           TO STDOUT [WITH(option[,...])]" query to execute
        :type query: str

        :return: response object
        :rtype: Response

        :raise CartoException:
        """
        url = self.api_url + '/copyto'
        params = {'api_key': self.api_key, 'q': query}

        try:
            response = self.client.send(url,
                                        http_method='GET',
                                        params=params,
                                        stream=True)
            response.raise_for_status()
        except HTTPError as e:
            if 400 <= response.status_code < 500:
                # Client error, provide better reason
                reason = response.json()['error'][0]
                error_msg = u'%s Client Error: %s' % (response.status_code,
                                                      reason)
                raise CartoException(error_msg)
            else:
                raise CartoException(e)
        except Exception as e:
            raise CartoException(e)

        return response

    def copyto_file_object(self, query, file_object):
        """
        Gets data from a table into a writable file object

        :param query: The "COPY { table_name [(column_name[, ...])] | (query) }
                           TO STDOUT [WITH(option[,...])]" query to execute
        :type query: str

        :param file_object: A file-like object.
                            Normally the return value of open('file.ext', 'wb')
        :type file_object: file

        :raise CartoException:
        """
        response = self.copyto(query)
        for block in response.iter_content(DEFAULT_CHUNK_SIZE):
            file_object.write(block)

    def copyto_file_path(self, query, path, append=False):
        """
        Gets data from a table into a writable file

        :param query: The "COPY { table_name [(column_name[, ...])] | (query) }
                           TO STDOUT [WITH(option[,...])]" query to execute
        :type query: str

        :param path: A path to a writable file
        :type path: str

        :param append: Whether to append or not if the file already exists
                       Default value is False
        :type append: bool

        :raise CartoException:
        """
        file_mode = 'wb' if not append else 'ab'
        with open(path, file_mode) as f:
            self.copyto_file_object(query, f)



"""
Module for working with CARTO datasets

.. module:: carto.datasets
   :platform: Unix, Windows
   :synopsis: Module for working with CARTO datasets

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

import time
import json
from gettext import gettext as _

from pyrestcli.fields import IntegerField, CharField, DateTimeField, \
    BooleanField

from .exceptions import CartoException
from .file_import import FileImportJobManager
from .resources import WarnResource
from .sync_tables import SyncTableJobManager
from .tables import TableManager
from .fields import TableField, UserField, PermissionField
from .paginators import CartoPaginator
from .resources import Manager


API_VERSION = "v1"
API_ENDPOINT = "api/{api_version}/viz/"

MAX_NUMBER_OF_RETRIES = 60
INTERVAL_BETWEEN_RETRIES_S = 10


class Dataset(WarnResource):
    """
    Represents a dataset in CARTO. Typically, that means there is a table in
    the PostgreSQL server associated to this object.

    .. warning:: Non-public API. It may change with no previous notice
    """
    active_child = None
    active_layer_id = CharField()
    attributions = CharField()
    auth_tokens = CharField(many=True)
    children = None
    created_at = DateTimeField()
    connector = None
    description = CharField()
    display_name = CharField()
    external_source = None
    id = CharField()
    kind = CharField()
    license = CharField()
    liked = BooleanField()
    likes = IntegerField()
    locked = BooleanField()
    map_id = CharField()
    name = CharField()
    next_id = None
    parent_id = CharField()
    permission = PermissionField()
    prev_id = None
    privacy = CharField()
    source = None
    stats = DateTimeField(many=True)
    synchronization = None
    table = TableField()
    tags = CharField(many=True)
    title = CharField()
    transition_options = None
    type = CharField()
    updated_at = DateTimeField()
    url = CharField()
    uses_builder_features = BooleanField()
    user = UserField()

    class Meta:
        collection_endpoint = API_ENDPOINT.format(api_version=API_VERSION)
        name_field = "name"


class DatasetManager(Manager):
    """
    Manager for the Dataset class.

    .. warning:: Non-public API. It may change with no previous notice
    """
    resource_class = Dataset
    json_collection_attribute = "visualizations"
    paginator_class = CartoPaginator

    def send(self, url, http_method, **client_args):
        """
        Sends an API request, taking into account that datasets are part of
        the visualization endpoint.

        :param url: Endpoint URL
        :param http_method: The method used to make the request to the API
        :param client_args: Arguments to be sent to the auth client
        :type url: str
        :type http_method: str
        :type client_args: kwargs

        :return: A request response object

        :raise: CartoException
        """
        try:
            client_args = client_args or {}

            if "params" not in client_args:
                client_args["params"] = {}
            client_args["params"].update({"type": "table",
                                          "exclude_shared": "true"})

            return super(DatasetManager, self).send(url,
                                                    http_method,
                                                    **client_args)
        except Exception as e:
            raise CartoException(e)

    def is_sync_table(self, archive, interval, **import_args):
        """
        Checks if this is a request for a sync dataset.

        The condition for creating a sync dataset is to provide a URL or a
        connection to an external database and an interval in seconds

        :param archive: URL to the file (both remote URLs or local paths are
                    supported) or StringIO object
        :param interval: Interval in seconds.
        :param import_args: Connection parameters for an external database
        :type url: str
        :type interval: int
        :type import_args: kwargs

        :return: True if it is a sync dataset

        """
        return (hasattr(archive, "startswith") and archive.startswith("http")
                or "connection" in import_args) \
            and interval is not None

    def create(self, archive, interval=None, **import_args):
        """
        Creating a table means uploading a file or setting up a sync table

        :param archive: URL to the file (both remote URLs or local paths are
                    supported) or StringIO object
        :param interval: Interval in seconds.
                        If not None, CARTO will try to set up a sync table
                        against the (remote) URL
        :param import_args: Arguments to be sent to the import job when run
        :type archive: str
        :type interval: int
        :type import_args: kwargs

        :return: New dataset object
        :rtype: Dataset

        :raise: CartoException
        """
        archive = archive.lower() if hasattr(archive, "lower") else archive

        if self.is_sync_table(archive, interval, **import_args):
            manager = SyncTableJobManager(self.client)
        else:
            manager = FileImportJobManager(self.client)

        import_job = manager.create(archive) if interval is None \
            else manager.create(archive, interval)
        import_job.run(**import_args)

        if import_job.get_id() is None:
            raise CartoException(_("Import API returned corrupt job details \
                                   when creating dataset"))

        import_job.refresh()

        count = 0
        while import_job.state in ("enqueued", "queued", "pending", "uploading",
                                   "unpacking", "importing", "guessing") \
            or (isinstance(manager, SyncTableJobManager)
                and import_job.state == "created"):
            if count >= MAX_NUMBER_OF_RETRIES:
                raise CartoException(_("Maximum number of retries exceeded \
                                       when polling the import API for \
                                       dataset creation"))
            time.sleep(INTERVAL_BETWEEN_RETRIES_S)
            import_job.refresh()
            count += 1

        if import_job.state == "failure":
            raise CartoException(_("Dataset creation was not successful \
                                   because of failed import (error: {error}")
                                 .format(error=json.dumps(
                                     import_job.get_error_text)))

        if (import_job.state != "complete" and import_job.state != "created"
            and import_job.state != "success") \
                or import_job.success is False:
            raise CartoException(_("Dataset creation was not successful \
                                   because of unknown import error"))

        if hasattr(import_job, "visualization_id") \
                and import_job.visualization_id is not None:
            visualization_id = import_job.visualization_id
        else:
            table = TableManager(self.client).get(import_job.table_id)
            visualization_id = table.table_visualization.get_id() \
                if table is not None else None

        try:
            return self.get(visualization_id) if visualization_id is not None \
                else None
        except AttributeError:
            raise CartoException(_("Dataset creation was not successful \
                                   because of unknown error"))



"""
Module for exporting visualizations

.. module:: carto.export
   :platform: Unix, Windows
   :synopsis: Module for exporting visualizations

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from pyrestcli.fields import CharField, DateTimeField

from .resources import WarnAsyncResource


API_VERSION = "v3"
API_ENDPOINT = 'api/{api_version}/visualization_exports/'


class ExportJob(WarnAsyncResource):
    """
    Equivalent to a .carto export in CARTO.

    Allows a CARTO export to be created using a visualization in the user's
    CARTO account

    .. warning:: Non-public API. It may change with no previous notice
    """
    id = CharField()
    visualization_id = CharField()
    user_id = CharField()
    state = CharField()
    url = CharField()
    created_at = DateTimeField()
    updated_at = DateTimeField()

    class Meta:
        collection_endpoint = API_ENDPOINT.format(api_version=API_VERSION)

    def __init__(self, client, visualization_id):
        """

        :param client: Client to make authorized requests
        :param visualization_id: The id of the visualization (or dataset!!!)
                                that will be exported
        :type client: :class:`carto.auth.APIKeyAuthClient`
        :type visualization_id: str

        :return:
        """

        self.visualization_id = visualization_id

        super(ExportJob, self).__init__(client)

    def run(self, **export_params):
        """
        Make the actual request to the Import API (exporting is part of the
        Import API).

        :param export_params: Any additional parameters to be sent to the
                                Import API
        :type export_params: kwargs

        :return:

        .. note:: The export is asynchronous, so you should take care of the progression, by calling the :func:`carto.resources.AsyncResource.refresh` method and check the export job :py:attr:`~state` attribute. See :func:`carto.visualizations.Visualization.export` method implementation for more details
        """
        export_params["visualization_id"] = self.visualization_id

        return super(ExportJob, self).run(params=export_params)



"""
Module for working with named and anonymous maps

.. module:: carto.maps
   :platform: Unix, Windows
   :synopsis: Module for working with named and anonymous maps

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from pyrestcli.resources import Manager, Resource

from .exceptions import CartoException

API_VERSION = "v1"
NAMED_API_ENDPOINT = "api/{api_version}/map/named/"
ANONYMOUS_API_ENDPOINT = "api/{api_version}/map/"


class BaseMap(Resource):
    """
    Base class for NamedMap and AnonymousMap
    """
    def __init__(self, auth_client):
        """
        Initializes a BaseMap instance
        :param auth_client: Auth client
        """
        super(BaseMap, self).__init__(auth_client)

    def get_tile_url(self, x, y, z, layer_id=None, feature_id=None,
                     filter=None, extension="png"):
        """
        Prepares a URL to get data (raster or vector) from a NamedMap or
        AnonymousMap

        :param x: The x tile
        :param y: The y tile
        :param z: The zoom level
        :param layer_id: Can be a number (referring to the # layer of your  \
                        map), all layers of your map, or a list of layers.
                        To show just the basemap layer, enter the value 0
                        To show the first layer, enter the value 1
                        To show all layers, enter the value 'all'
                        To show a list of layers, enter the comma separated \
                        layer value as '0,1,2'
        :param feature_id: The id of the feature
        :param filter: The filter to be applied to the layer
        :param extension: The format of the data to be retrieved: png, mvt, ...
        :type x: int
        :type y: int
        :type z: int
        :type layer_id: str
        :type feature_id: str
        :type filter: str
        :type extension: str

        :return: A URL to download data
        :rtype: str

        :raise: CartoException
        """
        base_url = self.client.base_url + self.Meta.collection_endpoint
        template_id = self.template_id if hasattr(self, 'template_id') \
            else self.layergroupid
        if layer_id is not None and feature_id is not None:
            url = urljoin(base_url,
                          "{template_id}/{layer}/attributes/{feature_id}"). \
                          format(template_id=template_id,
                                 layer=layer_id,
                                 feature_id=feature_id)
        elif layer_id is not None and filter is not None:
            url = urljoin(base_url,
                          "{template_id}/{filter}/{z}/{x}/{y}.{extension}"). \
                          format(template_id=template_id,
                                 filter=filter,
                                 z=z, x=x, y=y,
                                 extension=extension)
        elif layer_id is not None:
            url = urljoin(base_url,
                          "{template_id}/{layer}/{z}/{x}/{y}.{extension}"). \
                          format(template_id=template_id,
                                 layer=layer_id,
                                 z=z, x=x, y=y,
                                 extension=extension)
        else:
            url = urljoin(base_url, "{template_id}/{z}/{x}/{y}.{extension}"). \
                          format(
                               template_id=template_id,
                               z=z, x=x, y=y,
                               extension=extension)

        if hasattr(self, 'auth') and self.auth is not None \
           and len(self.auth['valid_tokens']) > 0:
            url = urljoin(url, "?auth_token={auth_token}"). \
                format(auth_token=self.auth['valid_tokens'][0])

        return url


class NamedMap(BaseMap):
    """
    Equivalent to creating a named map in CARTO.
    """
    class Meta:
        collection_endpoint = NAMED_API_ENDPOINT.format(
            api_version=API_VERSION)
        id_field = "template_id"
        name_field = "name"

    def __str__(self):
        try:
            return unicode(self.name).encode("utf-8")
        except AttributeError:
            return super(NamedMap, self).__repr__()

    def __init__(self, auth_client):
        """
        Initializes a NamedMap instance
        :param auth_client: Auth client
        """
        self.fields = ("version",
                       "name",
                       "auth",
                       "placeholders",
                       "layergroup",
                       "view")
        super(NamedMap, self).__init__(auth_client)

    def instantiate(self, params, auth=None):
        """
        Allows you to fetch the map tiles of a created map

        :param params: The json with the styling info for the named map
        :param auth: The auth client
        :type params: dict
        :type auth: :class:`carto.auth.APIKeyAuthClient`

        :return:

        :raise: CartoException
        """
        try:
            endpoint = (self.Meta.collection_endpoint
                        + "{template_id}"). \
                format(template_id=self.template_id)
            if (auth is not None):
                endpoint = (endpoint + "?auth_token={auth_token}"). \
                    format(auth_token=auth)

            self.send(endpoint, "POST", json=params)
        except Exception as e:
            raise CartoException(e)

    def update_from_dict(self, attribute_dict):
        """
        Method overriden from the base class

        """
        if 'template' in attribute_dict:
            self.update_from_dict(attribute_dict['template'])
            setattr(self,
                    self.Meta.id_field, attribute_dict['template']['name'])
            return
        try:
            for k, v in attribute_dict.items():
                setattr(self, k, v)
        except Exception:
            setattr(self, self.Meta.id_field, attribute_dict)


class AnonymousMap(BaseMap):
    """
    Equivalent to creating an anonymous map in CARTO.
    """
    class Meta:
        collection_endpoint = ANONYMOUS_API_ENDPOINT.format(
            api_version=API_VERSION)

    def instantiate(self, params):
        """
        Allows you to fetch the map tiles of a created map

        :param params: The json with the styling info for the named map
        :type params: dict

        :return:

        :raise: CartoException
        """
        try:
            self.send(self.Meta.collection_endpoint, "POST", json=params)
        except Exception as e:
            raise CartoException(e)

    def update_from_dict(self, attribute_dict):
        for k, v in attribute_dict.items():
            setattr(self, k, v)


class NamedMapManager(Manager):
    """
    Manager for the NamedMap class
    """
    resource_class = NamedMap
    json_collection_attribute = "template_ids"

    def create(self, **kwargs):
        """
        Creates a named map

        :param kwargs: Attributes for creating the named map. Specifically
                        an attribute `template` must contain the JSON object
                        defining the named map
        :type kwargs: kwargs

        :return: New named map object
        :rtype: NamedMap

        :raise: CartoException
        """
        resource = self.resource_class(self.client)
        resource.update_from_dict(kwargs['template'])
        resource.save(force_create=True)

        return resource



"""
Module for working with map visualizations

.. module:: carto.visualizations
   :platform: Unix, Windows
   :synopsis: Module for working with map visualizations

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

import time
from gettext import gettext as _

from pyrestcli.fields import IntegerField, CharField, DateTimeField, \
    BooleanField

from .exceptions import CartoException
from .fields import TableField
from .resources import Manager, WarnResource
from .paginators import CartoPaginator
from .export import ExportJob


API_VERSION = "v1"
API_ENDPOINT = "api/{api_version}/viz/"

MAX_NUMBER_OF_RETRIES = 30
INTERVAL_BETWEEN_RETRIES_S = 5


class Visualization(WarnResource):
    """
    Represents a map visualization in CARTO.

    .. warning:: Non-public API. It may change with no previous notice
    """
    active_child = None
    active_layer_id = CharField()
    attributions = None
    children = None
    created_at = DateTimeField()
    description = CharField()
    display_name = CharField()
    external_source = None
    id = CharField()
    kind = None
    license = None
    liked = BooleanField()
    likes = IntegerField()
    locked = BooleanField()
    map_id = CharField()
    name = CharField()
    next_id = None
    parent_id = None
    permission = None
    prev_id = None
    privacy = None
    source = None
    stats = None
    synchronization = None
    table = TableField()
    related_tables = TableField(many=True)
    tags = None
    title = CharField()
    transition_options = None
    type = None
    updated_at = DateTimeField()
    url = CharField()
    uses_builder_features = None

    class Meta:
        collection_endpoint = API_ENDPOINT.format(api_version=API_VERSION)
        name_field = "name"

    def export(self):
        """
        Make the actual request to the Import API (exporting is part of the
        Import API) to export a map visualization as a .carto file

        :return: A URL pointing to the .carto file
        :rtype: str

        :raise: CartoException

        .. warning:: Non-public API. It may change with no previous notice

        .. note:: The export is asynchronous, but this method waits for the export to complete. See `MAX_NUMBER_OF_RETRIES` and `INTERVAL_BETWEEN_RETRIES_S`
        """
        export_job = ExportJob(self.client, self.get_id())
        export_job.run()

        export_job.refresh()

        count = 0
        while export_job.state in ("exporting", "enqueued", "pending"):
            if count >= MAX_NUMBER_OF_RETRIES:
                raise CartoException(_("Maximum number of retries exceeded \
                                       when polling the import API for \
                                       visualization export"))
            time.sleep(INTERVAL_BETWEEN_RETRIES_S)
            export_job.refresh()
            count += 1

        if export_job.state == "failure":
            raise CartoException(_("Visualization export failed"))

        if (export_job.state != "complete" and export_job.state != "created"):
            raise CartoException(_("Unexpected problem on visualization export \
                                   (state: {state})").
                                 format(state=export_job.state))

        return export_job.url


class VisualizationManager(Manager):
    """
    Manager for the Visualization class.

    .. warning:: Non-public API. It may change with no previous notice
    """
    resource_class = Visualization
    json_collection_attribute = "visualizations"
    paginator_class = CartoPaginator

    def send(self, url, http_method, **client_args):
        """
        Sends API request, taking into account that visualizations are only a
        subset of the resources available at the visualization endpoint

        :param url: Endpoint URL
        :param http_method: The method used to make the request to the API
        :param client_args: Arguments to be sent to the auth client
        :type url: str
        :type http_method: str
        :type client_args: kwargs

        :return:

        :raise: CartoException
        """
        try:
            client_args.setdefault('params', {})
            client_args["params"].update({"type": "derived",
                                          "exclude_shared": "true"})

            return super(VisualizationManager, self).send(url,
                                                          http_method,
                                                          **client_args)
        except Exception as e:
            raise CartoException(e)

    def create(self, **kwargs):
        """
        Creating visualizations is better done by using the Maps API
        (named maps) or directly from your front end app if dealing with
        public datasets
        """
        pass



"""
Module for defining response objects

.. module:: carto.fields
   :platform: Unix, Windows
   :synopsis: Module for defining response objects

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from pyrestcli.fields import ResourceField


class VisualizationField(ResourceField):
    """
    :class:`carto.visualizations.Visualization`
    """
    value_class = "carto.visualizations.Visualization"


class TableField(ResourceField):
    """
    :class:`carto.tables.Table`
    """
    value_class = "carto.tables.Table"


class UserField(ResourceField):
    """
    :class:`carto.users.User`
    """
    value_class = "carto.users.User"


class EntityField(ResourceField):
    """
    :class:`carto.permissions.Entity`
    """
    value_class = "carto.permissions.Entity"


class PermissionField(ResourceField):
    """
    :class:`carto.permissions.Permission`
    """
    value_class = "carto.permissions.Permission"



# -*- coding: utf-8 -*-
"""
Module for authenticated access to CARTO's APIs

.. module:: carto.auth
   :platform: Unix, Windows
   :synopsis: Module for authenticated access to CARTO's APIs

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>
.. moduleauthor:: Juan Ignacio Sánchez <juanignaciosl@carto.com>


"""

from gettext import gettext as _
import re
import sys
import warnings
import pkg_resources

from pyrestcli.auth import BaseAuthClient, BasicAuthClient

from .exceptions import CartoException

if sys.version_info >= (3, 0):
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


class _UsernameGetter:
    def get_user_name(self, base_url):
        try:
            url_info = urlparse(base_url)
            # On-Prem:
            #   /user/<username>
            m = re.search('^/user/([^/]+)/.*$', url_info.path)
            if m is None:
                # Cloud personal account (org and standalone)
                # <username>.carto.com
                netloc = url_info.netloc
                if netloc.startswith('www.'):
                    netloc = netloc.split('www.')[1]
                m = re.search('^(.*?)\..*', netloc)
            return m.group(1)
        except Exception:
            raise CartoException(_("Could not find a valid user_name in the " +
                                   "base URL provided. Please check that the" +
                                   "URL is one of " +
                                   "'https://{user_name}.carto.com', " +
                                   "'https://carto.com/user/{user_name}' " +
                                   "or a similar one based on your domain"))


class _BaseUrlChecker:
    def check_base_url(self, base_url):
        if not base_url.startswith("https"):
            warnings.warn("You are using unencrypted API key \
                          authentication!!!")
        # Make sure there is a trailing / for urljoin
        if not base_url.endswith('/'):
            base_url += '/'

        return base_url


class _ClientIdentifier:

    CARTO_VERSION = pkg_resources.require('carto')[0].version

    def get_user_agent(self, name='carto-python-sdk'):
        return "{name}/{version}".format(
            name=name,
            version=self.CARTO_VERSION)

    def get_client_identifier(self, prefix='cps'):
        return "{prefix}-{version}".format(
            prefix=prefix,
            version=self.CARTO_VERSION)


class APIKeyAuthClient(_UsernameGetter, _BaseUrlChecker, _ClientIdentifier,
                       BaseAuthClient):
    """
    This class provides you with authenticated access to CARTO's APIs using
    your API key.

    You can find your API key by clicking on the API key section of the user
    dropdown menu
    """
    def __init__(self, base_url, api_key, organization=None, session=None):
        """
        Init method

        :param base_url: Base URL. API endpoint paths will always be relative
        to this URL
        :param api_key: API key
        :param organization: For enterprise users, organization user belongs to
        :param session: requests' session object
        :type api_key: str
        :type organization: str

        :return:
        """
        self.organization = organization
        self.api_key = api_key
        base_url = self.check_base_url(base_url)
        self.username = self.get_user_name(base_url)
        self.user_agent = self.get_user_agent()
        self.client_id = self.get_client_identifier()

        super(APIKeyAuthClient, self).__init__(base_url, session=session)

    def send(self, relative_path, http_method, **requests_args):
        """
        Makes an API-key-authorized request

        :param relative_path: URL path relative to self.base_url
        :param http_method: HTTP method
        :param requests_args: kwargs to be sent to requests
        :type relative_path: str
        :type http_method: str
        :type requests_args: kwargs

        :return:
            A request response object
        :raise:
            CartoException
        """
        try:
            http_method, requests_args = self.prepare_send(http_method, **requests_args)

            return super(APIKeyAuthClient, self).send(relative_path,
                                                      http_method,
                                                      **requests_args)
        except Exception as e:
            raise CartoException(e)

    def prepare_send(self, http_method, **requests_args):
        http_method = http_method.lower()
        params = {
            "api_key": self.api_key,
            "client": self.client_id
        }
        if (http_method in ['post', 'put']) and "json" in requests_args:
            requests_args["json"].update(params)
        else:
            if "params" not in requests_args:
                requests_args["params"] = {}
            requests_args["params"].update(params)

        if not requests_args.get('headers', None):
            requests_args['headers'] = {}
        requests_args['headers'].update({'User-Agent': self.user_agent})

        return http_method, requests_args


class NonVerifiedAPIKeyAuthClient(APIKeyAuthClient):
    """
    This class provides you with authenticated access to CARTO's APIs using
    your API key but avoids verifying SSL certificates. This is useful for onpremises instances of CARTO

    You can find your API key by clicking on the API key section of the user
    dropdown menu
    """
    def __init__(self, base_url, api_key, organization=None, session=None):
        """
        Init method

        :param base_url: Base URL. API endpoint paths will always be relative
        to this URL
        :param api_key: API key
        :param organization: For enterprise users, organization user belongs to
        :param session: requests' session object
        :type api_key: str
        :type organization: str

        :return:
        """
        super(NonVerifiedAPIKeyAuthClient, self).__init__(base_url, api_key, organization, session)

    def send(self, relative_path, http_method, **requests_args):
        """
        Makes an API-key-authorized request not verifying SSL certs

        :param relative_path: URL path relative to self.base_url
        :param http_method: HTTP method
        :param requests_args: kwargs to be sent to requests
        :type relative_path: str
        :type http_method: str
        :type requests_args: kwargs

        :return:
            A request response object
        :raise:
            CartoException
        """
        try:
            http_method, requests_args = self.prepare_send(http_method, **requests_args)
            requests_args["verify"] = False
            return super(APIKeyAuthClient, self).send(relative_path,
                                                      http_method,
                                                      **requests_args)
        except Exception as e:
            raise CartoException(e)


class AuthAPIClient(_UsernameGetter, _BaseUrlChecker, BasicAuthClient):
    """
    This class provides you with authenticated access to CARTO's APIs using
    your API key at Basic authentication header, as provided by Auth API.

    Auth API is still under development. You might want to take a look at
    APIKeyAuthClient for missing features or an stable API.

    You can find your API key by clicking on the API key section of the user
    dropdown menu
    """

    def __init__(self, base_url, api_key, organization=None, session=None):
        """
        Init method

        :param base_url: Base URL. API endpoint paths will always be relative
        to this URL
        :param api_key: API key
        :param organization: For enterprise users, organization user belongs to
        :param session: requests' session object
        :type api_key: str
        :type organization: str

        :return:
        """
        self.organization = organization
        self.api_key = api_key
        base_url = self.check_base_url(base_url)
        self.username = self.get_user_name(base_url)

        super(AuthAPIClient, self).__init__(self.username, api_key, base_url, session=session)

    def is_valid_api_key(self):
        """
        Checks validity. Right now, an API key is considered valid if it
        can list user API keys and the result contains that API key.
        This might change in the future.

        :return: True if the API key is considered valid for current user.
        """
        res = self.send('api/v3/api_keys', 'get')
        return \
            res.ok and \
            self.api_key in (ak['token'] for ak in res.json()['result'])



"""
Extensions for pyrestcli Resource and Manager classes

.. module:: carto.resources
   :platform: Unix, Windows
   :synopsis: Extensions for pyrestcli Resource and Manager classes

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

import warnings

from pyrestcli.resources import Resource, Manager as PyRestCliManager

from .exceptions import CartoException


class AsyncResource(Resource):
    def run(self, **client_params):
        """
        Actually creates the async job on the CARTO server


        :param client_params: To be send to the CARTO API. See CARTO's
                                documentation depending on the subclass
                                you are using
        :type client_params: kwargs


        :return:
        :raise: CartoException
        """
        try:
            self.send(self.get_collection_endpoint(),
                      http_method="POST",
                      **client_params)
        except Exception as e:
            raise CartoException(e)

    def refresh(self):
        """
        Updates the information of the async job against the CARTO server.
        After calling the :func:`refresh` method you should check the `state`
        attribute of your resource

        :return:
        """
        if self.get_resource_endpoint() is None:
            raise CartoException("Async job needs to be run or retrieved \
                                 first!")

        super(AsyncResource, self).refresh()


class WarnAsyncResource(AsyncResource):
    """
    AsyncResource class for resources that represent non-public CARTO APIs.
    You'll be warned not to used the in production environments
    """
    def __init__(self, auth_client, **kwargs):
        """
        Initializes the resource
        :param auth_client: Client to make (non)authorized requests
        :param kwargs: Initial value for attributes
        :return:
        """

        warnings.warn('This is part of a non-public CARTO API and may change in \
              the future. Take this into account if you are using \
              this in a production environment', FutureWarning)
        super(WarnAsyncResource, self).__init__(auth_client, **kwargs)


class WarnResource(Resource):
    """
    Resource class for resources that represent non-public CARTO APIs.
    You'll be warned not to used the in production environments
    """
    def __init__(self, auth_client, **kwargs):
        """
        Initializes the resource
        :param auth_client: Client to make (non)authorized requests
        :param kwargs: Initial value for attributes
        :return:
        """

        warnings.warn('This is part of a non-public CARTO API and may change in the future. Take this into account if you are using this in a production environment', FutureWarning)
        super(WarnResource, self).__init__(auth_client, **kwargs)


class Manager(PyRestCliManager):
    """
    Manager class for resources
    """
    def __init__(self, auth_client):
        """
        :param auth_client: Client to make (non)authorized requests

        :return:
        """
        self.paginator = self.paginator_class(self.json_collection_attribute,
                                              auth_client.base_url)
        super(PyRestCliManager, self).__init__(auth_client)



"""
Module for working with tables

.. module:: carto.tables
   :platform: Unix, Windows
   :synopsis: Module for working with tables

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from pyrestcli.fields import IntegerField, CharField, DateTimeField

from .fields import PermissionField, VisualizationField
from .paginators import CartoPaginator
from .resources import Manager, WarnResource


API_VERSION = "v1"
API_ENDPOINT = "api/{api_version}/tables/"


class Table(WarnResource):
    """
    Represents a table in CARTO. This is an internal data type. Both Table and
    TableManager are not meant to be used outside the SDK

    If you are looking to work with datasets / tables from outside the SDK,
    please look into the datasets.py file.

    .. warning:: Non-public API. It may change with no previous notice
    """
    id = CharField()
    name = CharField()
    privacy = CharField()
    permission = PermissionField()
    schema = None
    updated_at = DateTimeField()
    rows_counted = IntegerField()
    row_count = IntegerField()
    size = IntegerField()
    table_size = IntegerField()
    map_id = CharField()
    description = CharField()
    geometry_types = CharField(many=True)
    table_visualization = VisualizationField()
    dependent_visualizations = None
    non_dependent_visualizations = None
    synchronization = None

    class Meta:
        collection_endpoint = API_ENDPOINT.format(api_version=API_VERSION)
        name_field = "name"


class TableManager(Manager):
    """
    Manager for the Table class.

    .. warning:: Non-public API. It may change with no previous notice
    """
    resource_class = Table
    paginator_class = CartoPaginator



"""
Entity classes for defining permissions

.. module:: carto.permissions
   :platform: Unix, Windows
   :synopsis: Entity classes for defining permissions

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from pyrestcli.resources import Resource
from pyrestcli.fields import CharField, DateTimeField

from .fields import UserField, EntityField


PUBLIC = "PUBLIC"
PRIVATE = "PRIVATE"
LINK = "LINK"


class Entity(Resource):
    """
    Represents an entity in CARTO. This is an internal data type, with no
    specific API endpoints
    """
    id = CharField()
    type = CharField()


class Permission(Resource):
    """
    Represents a permission in CARTO. This is an internal data type, with no
    specific API endpoints
    """
    acl = None
    created_at = DateTimeField()
    entity = EntityField()
    id = CharField()
    owner = UserField()
    updated_at = DateTimeField()



"""
Module for carto-python exceptions definitions

.. module:: carto.exceptions
   :platform: Unix, Windows
   :synopsis: Module for carto-python exceptions definitions

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""


class CartoException(Exception):
    """
    Any Exception produced by carto-python should be wrapped around this class
    """
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

try:
    with open('requirements.txt') as f:
        required = f.read().splitlines()
except:
    required = ['requests>=2.7.0', 'pyrestcli>=0.6.4']

try:
    with open('test_requirements.txt') as f:
        test_required = f.read().splitlines()
except:
    pass

setup(name="carto",
      author="Daniel Carrión",
      author_email="daniel@carto.com",
      description="SDK around CARTO's APIs",
      version="1.3.1",
      url="https://github.com/CartoDB/carto-python",
      install_requires=required,
      packages=["carto"])



"""
Used internally to retrieve results paginated

.. module:: carto.paginators
   :platform: Unix, Windows
   :synopsis: Used internally to retrieve results paginated

.. moduleauthor:: Daniel Carrion <daniel@carto.com>
.. moduleauthor:: Alberto Romeu <alrocar@carto.com>


"""

from pyrestcli.paginators import Paginator


class CartoPaginator(Paginator):
    """
    Used internally to retrieve results paginated

    """
    def __init__(self, json_collection_attribute, base_url, params=None):
        self.json_collection_attribute = json_collection_attribute

        super(CartoPaginator, self).__init__(base_url, params)

    def get_urls(self, initial_url):
        self.url = initial_url
        self.total_count = 0
        self.page = 1

        while self.url is not None:
            yield self.url, {"page": self.page}

    def process_response(self, response):
        response_json = response.json()
        self.total_count += len(response_json[self.json_collection_attribute])

        try:
            total_count_from_api = int(response_json["total_entries"])
        except KeyError:
            total_count_from_api = int(response_json["total_user_entries"])

        if self.total_count < total_count_from_api:
            self.page += 1
        else:
            self.url = None

        return response
